from aiohttp import WSMsgType
import json


class WebSocketsAPI:
    '''
    WebSockets API endpoint for streaming data.
    '''
    def __init__(self, api):
        self.api = api

    async def connect(self, streaming_url=None):
        '''
        Create WebSocket connection to streaming API.
        '''
        path = '/api/v1/streaming'
        ws = await self.api._websocket(path, streaming_url=streaming_url)
        return WebSocket(ws)


class WebSocket:

    def __init__(self, ws):
        self._ws = ws

    async def __aiter__(self):
        async for msg in self._ws:
            if msg.type == WSMsgType.TEXT:
                event = json.loads(msg.data)
                if 'payload' in event:
                    event['data'] = json.loads(event['payload'])
                    del event['payload']
                yield event
            elif msg.type == WSMsgType.ERROR:
                break

    async def subscribe(self, stream, **kwargs):
        kwargs['type'] = 'subscribe'
        kwargs['stream'] = stream
        await self._ws.send_str(json.dumps(kwargs))



class AccountsAPI:
    '''
    Account related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def verify_credentials(self):
        '''
        Test to make sure that the user token works.
        https://docs.joinmastodon.org/methods/accounts/#verify_credentials
        '''
        path = '/api/v1/accounts/verify_credentials'
        return await self.api._get(path)

    async def get(self, id):
        '''
        View information about a profile.
        https://docs.joinmastodon.org/methods/accounts/#get
        '''
        path = f'/api/v1/accounts/{id}'
        return await self.api._get(path)

    async def statuses(self, id, **query):
        '''
        Statuses posted to the given account.
        https://docs.joinmastodon.org/methods/accounts/#statuses
        '''
        path = f'/api/v1/accounts/{id}/statuses'
        return await self.api._get(path, query=query)

    async def followers(self, id, **query):
        '''
        Accounts which follow given account, if network isn't hidden.
        https://docs.joinmastodon.org/methods/accounts/#followers
        '''
        path = f'/api/v1/accounts/{id}/followers'
        return await self.api._get(path, query=query)

    async def following(self, id, **query):
        '''
        Accounts which the given account follows, if network isn't hidden.
        https://docs.joinmastodon.org/methods/accounts/#following
        '''
        path = f'/api/v1/accounts/{id}/following'
        return await self.api._get(path, query=query)

    async def featured_tags(self, id):
        '''
        Tags featured by this account.
        https://docs.joinmastodon.org/methods/accounts/#featured_tags
        '''
        path = f'/api/v1/accounts/{id}/featured_tags'
        return await self.api._get(path)

    async def lists(self, id):
        '''
        User lists that you have added this account to.
        https://docs.joinmastodon.org/methods/accounts/#lists
        '''
        path = f'/api/v1/accounts/{id}/lists'
        return await self.api._get(path)

    async def follow(self, id, **body):
        '''
        Follow the given account.
        Can also update whether to show reblogs or enable notifications.
        https://docs.joinmastodon.org/methods/accounts/#follow
        '''
        path = f'/api/v1/accounts/{id}/follow'
        return await self.api._post(path, body=body)

    async def unfollow(self, id):
        '''
        Unfollow the given account.
        https://docs.joinmastodon.org/methods/accounts/#unfollow
        '''
        path = f'/api/v1/accounts/{id}/unfollow'
        return await self.api._post(path)

    async def remove_from_followers(self, id):
        '''
        Remove the given account from your followers.
        https://docs.joinmastodon.org/methods/accounts/#remove_from_followers
        '''
        path = f'/api/v1/accounts/{id}/remove_from_followers'
        return await self.api._post(path)

    async def block(self, id):
        '''
        Block the given account.
        https://docs.joinmastodon.org/methods/accounts/#block
        '''
        path = f'/api/v1/accounts/{id}/block'
        return await self.api._post(path)

    async def unblock(self, id):
        '''
        Unblock the given account.
        https://docs.joinmastodon.org/methods/accounts/#unblock
        '''
        path = f'/api/v1/accounts/{id}/unblock'
        return await self.api._post(path)

    async def mute(self, id):
        '''
        Mute the given account.
        https://docs.joinmastodon.org/methods/accounts/#mute
        '''
        path = f'/api/v1/accounts/{id}/mute'
        return await self.api._post(path)

    async def unmute(self, id):
        '''
        Unmute the given account.
        https://docs.joinmastodon.org/methods/accounts/#unmute
        '''
        path = f'/api/v1/accounts/{id}/unmute'
        return await self.api._post(path)

    async def pin(self, id):
        '''
        Add the given account to the user’s featured profiles.
        https://docs.joinmastodon.org/methods/accounts/#pin
        '''
        path = f'/api/v1/accounts/{id}/pin'
        return await self.api._post(path)

    async def unpin(self, id):
        '''
        Remove the given account from the user’s featured profiles.
        https://docs.joinmastodon.org/methods/accounts/#unpin
        '''
        path = f'/api/v1/accounts/{id}/unpin'
        return await self.api._post(path)

    async def note(self, id, text):
        '''
        Sets a private note on a user.
        https://docs.joinmastodon.org/methods/accounts/#note
        '''
        path = f'/api/v1/accounts/{id}/note'
        body = {'comment': text}
        return await self.api._post(path, body=body)

    async def relationships(self, *ids):
        '''
        Find out whether a given account is followed, blocked, muted, etc.
        https://docs.joinmastodon.org/methods/accounts/#relationships
        '''
        path = '/api/v1/accounts/relationships'
        query = '&'.join(f'id[]={id}' for id in ids)
        return await self.api._get(path + '?' + query)

    async def familiar_followers(self, *ids):
        '''
        Obtain a list of all accounts that follow a given account,
        filtered for accounts you follow.
        https://docs.joinmastodon.org/methods/accounts/#familiar_followers
        '''
        path = '/api/v1/accounts/familiar_followers'
        query = '&'.join(f'id[]={id}' for id in ids)
        return await self.api._get(path + '?' + query)

    async def search(self, q, **query):
        '''
        Search for matching accounts by username or display name.
        https://docs.joinmastodon.org/methods/accounts/#search
        '''
        path = '/api/v1/accounts/search'
        query['q'] = q
        return await self.api._get(path, query=query)

    async def lookup(self, acct):
        '''
        Quickly lookup a username to see if it is available,
        skipping WebFinger resolution.
        https://docs.joinmastodon.org/methods/accounts/#lookup
        '''
        path = '/api/v1/accounts/lookup'
        return await self.api._get(path, query={'acct': acct})



class ConversationsAPI:
    '''
    Conversation related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def all(self, **query):
        '''
        View all conversations.
        https://docs.joinmastodon.org/methods/conversations/#get
        '''
        path = '/api/v1/conversations'
        return await self.api._get(path, query=query)

    async def remove(self, id):
        '''
        Removes a conversation from your list of conversations.
        https://docs.joinmastodon.org/methods/conversations/#delete
        '''
        path = f'/api/v1/conversations/{id}'
        return await self.api._delete(path)

    async def read(self, id):
        '''
        Mark a conversation as read.
        https://docs.joinmastodon.org/methods/conversations/#read
        '''
        path = f'/api/v1/conversations/{id}/read'
        return await self.api._post(path)



class StatusesAPI:
    '''
    Status related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def post(self, status, **query):
        '''
        Publish a status with the given parameters.
        https://docs.joinmastodon.org/methods/statuses/#create
        '''
        path = '/api/v1/statuses'
        query['status'] = status
        return await self.api._post(path, query=query)

    async def get(self, id):
        '''
        Obtain information about a status.
        https://docs.joinmastodon.org/methods/statuses/#get
        '''
        path = f'/api/v1/statuses/{id}'
        return await self.api._get(path)

    async def delete(self, id):
        '''
        Delete one of your own statuses.
        https://docs.joinmastodon.org/methods/statuses/#delete
        '''
        path = f'/api/v1/statuses/{id}'
        return await self.api._delete(path)

    async def context(self, id):
        '''
        View statuses above and below this status in the thread.
        https://docs.joinmastodon.org/methods/statuses/#context
        '''
        path = f'/api/v1/statuses/{id}/context'
        return await self.api._get(path)

    async def get_boosts(self, id, **query):
        '''
        View who boosted a given status.
        https://docs.joinmastodon.org/methods/statuses/#reblogged_by
        '''
        path = f'/api/v1/statuses/{id}/reblogged_by'
        return await self.api._get(path, query=query)

    async def get_favourites(self, id, **query):
        '''
        View who favourited a given status.
        https://docs.joinmastodon.org/methods/statuses/#favourited_by
        '''
        path = f'/api/v1/statuses/{id}/favourited_by'
        return await self.api._get(path, query=query)

    async def favourite(self, id):
        '''
        Add a status to your favourites list.
        https://docs.joinmastodon.org/methods/statuses/#favourite
        '''
        path = f'/api/v1/statuses/{id}/favourite'
        return await self.api._post(path)

    async def unfavourite(self, id):
        '''
        Remove a status from your favourites list.
        https://docs.joinmastodon.org/methods/statuses/#unfavourite
        '''
        path = f'/api/v1/statuses/{id}/unfavourite'
        return await self.api._post(path)

    async def boost(self, id):
        '''
        Reshare a status on your own profile.
        https://docs.joinmastodon.org/methods/statuses/#boost
        '''
        path = f'/api/v1/statuses/{id}/reblog'
        return await self.api._post(path)

    async def unboost(self, id):
        '''
        Undo a reshare of a status.
        https://docs.joinmastodon.org/methods/statuses/#unreblog
        '''
        path = f'/api/v1/statuses/{id}/unreblog'
        return await self.api._post(path)

    async def bookmark(self, id):
        '''
        Privately bookmark a status.
        https://docs.joinmastodon.org/methods/statuses/#bookmark
        '''
        path = f'/api/v1/statuses/{id}/bookmark'
        return await self.api._post(path)

    async def unbookmark(self, id):
        '''
        Remove a status from your private bookmarks.
        https://docs.joinmastodon.org/methods/statuses/#unbookmark
        '''
        path = f'/api/v1/statuses/{id}/unbookmark'
        return await self.api._post(path)

    async def mute(self, id):
        '''
        Do not receive notifications for thread this status is part of.
        Must be a thread in which you are a participant.
        https://docs.joinmastodon.org/methods/statuses/#mute
        '''
        path = f'/api/v1/statuses/{id}/mute'
        return await self.api._post(path)

    async def unmute(self, id):
        '''
        Start receiving notifications again for thread this status is part of.
        https://docs.joinmastodon.org/methods/statuses/#unmute
        '''
        path = f'/api/v1/statuses/{id}/unmute'
        return await self.api._post(path)

    async def pin(self, id):
        '''
        Feature one of your own public statuses at the top of your profile.
        https://docs.joinmastodon.org/methods/statuses/#pin
        '''
        path = f'/api/v1/statuses/{id}/pin'
        return await self.api._post(path)

    async def unpin(self, id):
        '''
        Unfeature a status from the top of your profile.
        https://docs.joinmastodon.org/methods/statuses/#unpin
        '''
        path = f'/api/v1/statuses/{id}/unpin'
        return await self.api._post(path)

    async def edit(self, id, **body):
        '''
        Edit status to change text, sensitivity, media attachments, or poll.
        Note that editing a poll's options will reset the votes.
        https://docs.joinmastodon.org/methods/statuses/#edit
        '''
        path = f'/api/v1/statuses/{id}'
        return await self.api._put(path, body=body)

    async def get_history(self, id):
        '''
        Get all known versions of status, including initial and current states.
        https://docs.joinmastodon.org/methods/statuses/#history
        '''
        path = f'/api/v1/statuses/{id}/history'
        return await self.api._get(path)

    async def get_source(self, id):
        '''
        Obtain the source properties for a status so that it can be edited.
        https://docs.joinmastodon.org/methods/statuses/#source
        '''
        path = f'/api/v1/statuses/{id}/source'
        return await self.api._get(path)



from aiohttp import ClientSession

from .accounts import AccountsAPI
from .blocks import BlocksAPI
from .bookmarks import BookmarksAPI
from .conversations import ConversationsAPI
from .domain_blocks import DomainBlocksAPI
from .favourites import FavouritesAPI
from .instance import InstanceAPI
from .lists import ListsAPI
from .mutes import MutesAPI
from .notifications import NotificationsAPI
from .statuses import StatusesAPI
from .timelines import TimelinesAPI
from .trends import TrendsAPI
from .websockets import WebSocketsAPI


class MastodonAPI:
    '''
    Minimal asynchronous wrapper for Mastodon API.
    '''
    def __init__(self, url, ssl=None):
        self.url = _normalize_url(url)
        # SSL context (default context used if None)
        self.ssl = ssl
        self.session = ClientSession()
        self.accounts = AccountsAPI(self)
        self.blocks = BlocksAPI(self)
        self.bookmarks = BookmarksAPI(self)
        self.conversations = ConversationsAPI(self)
        self.domain_blocks = DomainBlocksAPI(self)
        self.favourites = FavouritesAPI(self)
        self.instance = InstanceAPI(self)
        self.lists = ListsAPI(self)
        self.mutes = MutesAPI(self)
        self.notifications = NotificationsAPI(self)
        self.statuses = StatusesAPI(self)
        self.timelines = TimelinesAPI(self)
        self.trends = TrendsAPI(self)
        self.websockets = WebSocketsAPI(self)

    def set_access_token(self, access_token):
        '''
        Set OAuth access token.
        '''
        self.session.headers['Authorization'] = f'Bearer {access_token}'

    async def _request(self, method, path, query=None, body=None):
        '''
        Perform HTTP request using method for API path and return parsed JSON.
        '''
        url = self.url + path
        kwargs = {}
        if self.ssl is not None:
            kwargs['ssl'] = self.ssl
        if query is not None:
            kwargs['params'] = query
        if body is not None:
            kwargs['json'] = body
        async with self.session.request(method, url, **kwargs) as resp:
            return await resp.json()

    async def _get(self, path, **kwargs):
        '''
        Perform GET request for API path and return parsed JSON.
        '''
        return await self._request('GET', path, **kwargs)

    async def _post(self, path, **kwargs):
        '''
        Perform POST request for API path and return parsed JSON.
        '''
        return await self._request('POST', path, **kwargs)

    async def _put(self, path, **kwargs):
        '''
        Perform PUT request for API path and return parsed JSON.
        '''
        return await self._request('PUT', path, **kwargs)

    async def _delete(self, path, **kwargs):
        '''
        Perform DELETE request for API path and return parsed JSON.
        '''
        return await self._request('DELETE', path, **kwargs)

    async def _websocket(self, path, streaming_url=None):
        '''
        Perform a WebSocket request for API path and return connection.
        '''
        if streaming_url is None:
            _, streaming_url = self.url.split('://', maxsplit=1)
            streaming_url = 'wss://' + streaming_url
        url = streaming_url + path
        return await self.session.ws_connect(url, ssl=self.ssl)

    async def close(self):
        await self.session.close()

    async def __aenter__(self, *args):
        return self

    async def __aexit__(self, *args):
        await self.close()


def _normalize_url(url):
    url = url.rstrip('/')
    if url.startswith('http://'):
        return url
    if url.startswith('https://'):
        return url
    return 'https://' + url



class DomainBlocksAPI:
    '''
    Domain blocks related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def all(self, **query):
        '''
        View domains the user has blocked.
        https://docs.joinmastodon.org/methods/domain_blocks/#get
        '''
        path = '/api/v1/domain_blocks'
        return await self.api._get(path, query=query)

    async def block(self, domain, **body):
        '''
        Block a domain.
        https://docs.joinmastodon.org/methods/domain_blocks/#block
        '''
        path = '/api/v1/domain_blocks'
        body['domain'] = domain
        return await self.api._post(path, body=body)

    async def unblock(self, domain, **body):
        '''
        Remove a domain block.
        https://docs.joinmastodon.org/methods/domain_blocks/#unblock
        '''
        path = '/api/v1/domain_blocks'
        body['domain'] = domain
        return await self.api._delete(path, body=body)



class InstanceAPI:
    '''
    Instance related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def info(self):
        '''
        Obtain general information about the server.
        https://docs.joinmastodon.org/methods/instance/#v2
        '''
        path = '/api/v2/instance'
        return await self.api._get(path)

    async def peers(self):
        '''
        Domains that this instance is aware of.
        https://docs.joinmastodon.org/methods/instance/#peers
        '''
        path = '/api/v1/instance/peers'
        return await self.api._get(path)

    async def activity(self):
        '''
        Instance activity over the last 3 months, binned weekly.
        https://docs.joinmastodon.org/methods/instance/#activity
        '''
        path = '/api/v1/instance/activity'
        return await self.api._get(path)

    async def rules(self):
        '''
        Rules that the users of this service should follow.
        https://docs.joinmastodon.org/methods/instance/#rules
        '''
        path = '/api/v1/instance/rules'
        return await self.api._get(path)

    async def domain_blocks(self):
        '''
        Obtain a list of domains that have been blocked.
        https://docs.joinmastodon.org/methods/instance/#domain_blocks
        '''
        path = '/api/v1/instance/domain_blocks'
        return await self.api._get(path)

    async def extended_description(self):
        '''
        Obtain an extended description of this server.
        https://docs.joinmastodon.org/methods/instance/#extended_description
        '''
        path = '/api/v1/instance/extended_description'
        return await self.api._get(path)

    async def directory(self, **query):
        '''
        List accounts visible in the directory.
        https://docs.joinmastodon.org/methods/directory/#get
        '''
        path = '/api/v1/directory'
        return await self.api._get(path, query=query)

    async def custom_emojis(self):
        '''
        Returns custom emojis that are available on the server.
        https://docs.joinmastodon.org/methods/custom_emojis/#get
        '''
        path = '/api/v1/custom_emojis'
        return await self.api._get(path)

    async def search(self, q, **query):
        '''
        Search for content in accounts, statuses and hashtags.
        https://docs.joinmastodon.org/methods/search/#v2
        '''
        path = '/api/v2/search'
        query['q'] = q
        return await self.api._get(path, query=query)



class BlocksAPI:
    '''
    Blocks related API endpoints.
    '''
    def __init__(self, api):
        self.api = api

    async def all(self, **query):
        '''
        View blocked users.
        https://docs.joinmastodon.org/methods/blocks/#get
        '''
        path = '/api/v1/blocks'
        return await self.api._get(path, query=query)
